#ifndef __SOPHIACRADLE_SANGI_HPP
#define __SOPHIACRADLE_SANGI_HPP

#include <SophiaFramework.hpp>
#include "Expression.hpp"
#include "Stack.hpp"
#include "SuperMath.hpp"

// t|[h\LvZNX
// o[W̒`
template <typename T>
class Sangi : public static_exception<SFCError> {
SFMSEALCOPY(Sangi)
private:
    Stack<T> _stack;
    T _result;

public:
    explicit Sangi(Void);
    virtual ~Sangi(Void);
    T Solve(const Expression<T> &expression);
    T GetResult(Void) const;
private:
    Void CheckParamCount(UInt16 count) const static_throws;
    Void Plus(Void) static_throws;
    Void Minus(Void) static_throws;
    Void Add(Void) static_throws;
    Void Sub(Void) static_throws;
    Void Mul(Void) static_throws;
    Void Div(Void) static_throws;
    Void Mod(Void) static_throws;
    Void And(Void) static_throws;
    Void Or(Void) static_throws;
    Void Xor(Void) static_throws;
    Void Not(Void) static_throws;
};

template <typename T>
inline Sangi<T>::Sangi(Void) : _result(0)
{
}

template <typename T>
inline Sangi<T>::~Sangi(Void)
{
    _stack.Clear();
}

// vZ
template <typename T>
inline T Sangi<T>::Solve(const Expression<T> &expression)
{
    SInt32 i;
    SInt32Const size(expression.GetSize());
    ExpressionElement<T> element;
    UInt16 type;

    _stack.Clear();
    _result = 0;
    errno = 0;
    if (size == 0) {
        return 0;
    }
    for (i = 0; i < size; ++i) {
        element = expression.Get(i);
        type = element.GetType();
        switch (type) {
            case ExpressionElement<T>::TYPE_VALUE:
                static_throw(_stack.Push(element.GetValue()));
                break;
            case ExpressionElement<T>::TYPE_PLUS:
                Plus();
                break;
            case ExpressionElement<T>::TYPE_MINUS:
                Minus();
                break;
            case ExpressionElement<T>::TYPE_ADD:
                Add();
                break;
            case ExpressionElement<T>::TYPE_SUB:
                Sub();
                break;
            case ExpressionElement<T>::TYPE_MUL:
                Mul();
                break;
            case ExpressionElement<T>::TYPE_DIV:
                Div();
                break;
            case ExpressionElement<T>::TYPE_MOD:
                Mod();
                break;
            case ExpressionElement<T>::TYPE_AND:
                And();
                break;
            case ExpressionElement<T>::TYPE_OR:
                Or();
                break;
            case ExpressionElement<T>::TYPE_XOR:
                Xor();
                break;
            case ExpressionElement<T>::TYPE_NOT:
                Not();
                break;
            default:
                static_throw(Calculator::ERR_SYNTAX);
                break;
        }
        if (errno != 0) {
            static_throw(Calculator::ERR_DOMAIN);
        }
        if (static_catch() != SFERR_NO_ERROR) {
            break;
        }
    }
    if (static_catch() == SFERR_NO_ERROR && _stack.GetSize() != 1) {
        static_throw(Calculator::ERR_SYNTAX);
    }
    if (static_catch() == SFERR_NO_ERROR) {
        _result = _stack.Pop();
    }
    return _result;
}

template <typename T>
inline Void Sangi<T>::CheckParamCount(UInt16 count) const static_throws
{
    if (_stack.GetSize() < count) {
        static_throw(Calculator::ERR_SYNTAX);
    }
}

template <typename T>
inline Void Sangi<T>::Plus(Void) static_throws
{
    CheckParamCount(1);
}

template <typename T>
inline Void Sangi<T>::Minus(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param(_stack.Pop());
    static_throw(_stack.Push(0 - param));
}

template <typename T>
inline Void Sangi<T>::Add(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] + param[1]));
}

template <typename T>
inline Void Sangi<T>::Sub(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] - param[1]));
}

template <typename T>
inline Void Sangi<T>::Mul(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] * param[1]));
}

template <typename T>
inline Void Sangi<T>::Div(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    if (param[1] == 0) {
        static_throw(Calculator::ERR_ZERO_DIVISION);
    }
    else {
        static_throw(_stack.Push(param[0] / param[1]));
    }
}

template <typename T>
inline Void Sangi<T>::Mod(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    if (param[1] == 0) {
        static_throw(Calculator::ERR_ZERO_DIVISION);
    }
    else {
        static_throw(_stack.Push(param[0] % param[1]));
    }
}

template <typename T>
inline Void Sangi<T>::And(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] & param[1]));
}

template <typename T>
inline Void Sangi<T>::Or(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] | param[1]));
}

template <typename T>
inline Void Sangi<T>::Xor(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] ^ param[1]));
}

template <typename T>
inline Void Sangi<T>::Not(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    T param(_stack.Pop());
    static_throw(_stack.Push(~param));
}

template <typename T>
inline T Sangi<T>::GetResult(Void) const
{
    return _result;
}

// t|[h\LvZNX
// o[W̒`
template < >
class Sangi<Float64> : public static_exception<SFCError> {
SFMSEALCOPY(Sangi)
public:
    enum AngleMode {
        ANGLEMODE_RAD = 0,          // WA
        ANGLEMODE_DEG,              // x
        ANGLEMODE_GRD               // OWA
    };
private:
    Stack<Float64> _stack;
    Float64 _result;
    AngleMode _angle;

public:
    explicit Sangi(Void);
    explicit Sangi(AngleMode mode);
    virtual ~Sangi(Void);
    Void SetAngleMode(AngleMode mode);
    Float64 Solve(const Expression<Float64> &expression);
    Float64 GetResult(Void) const;
private:
    Void CheckParamCount(UInt16 count) const static_throws;
    Void Plus(Void) static_throws;
    Void Minus(Void) static_throws;
    Void Add(Void) static_throws;
    Void Sub(Void) static_throws;
    Void Mul(Void) static_throws;
    Void Div(Void) static_throws;
    Void Mod(Void) static_throws;
    Void Factorial(Void) static_throws;
    Void DoubleFactorial(Void) static_throws;
    Void Pow(Void) static_throws;
    Void Exp(Void) static_throws;
    Void Ln(Void) static_throws;
    Void Log(Void) static_throws;
    Void Sqrt(Void) static_throws;
    Void Cbrt(Void) static_throws;
    Void Sin(Void) static_throws;
    Void Cos(Void) static_throws;
    Void Tan(Void) static_throws;
    Void Asin(Void) static_throws;
    Void Acos(Void) static_throws;
    Void Atan(Void) static_throws;
    Void Sinh(Void) static_throws;
    Void Cosh(Void) static_throws;
    Void Tanh(Void) static_throws;
    Void Asinh(Void) static_throws;
    Void Acosh(Void) static_throws;
    Void Atanh(Void) static_throws;
    Void Abs(Void) static_throws;
    Void Int(Void) static_throws;
    Void Fix(Void) static_throws;
    Void Gamma(Void) static_throws;
    Void nPr(Void) static_throws;
    Void nCr(Void) static_throws;
    Void Frac(Void) static_throws;
    Void Jn(Void) static_throws;
    Void Yn(Void) static_throws;
};

inline Sangi<Float64>::Sangi(Void) : _result(0), _angle(ANGLEMODE_RAD)
{
}

inline Sangi<Float64>::Sangi(Sangi<Float64>::AngleMode mode) : _result(0), _angle(mode)
{
}

inline Sangi<Float64>::~Sangi(Void)
{
    _stack.Clear();
}

inline Void Sangi<Float64>::SetAngleMode(Sangi<Float64>::AngleMode mode)
{
    _angle = mode;
}

// vZ
inline Float64 Sangi<Float64>::Solve(const Expression<Float64> &expression)
{
    SInt32 i;
    SInt32Const size(expression.GetSize());
    ExpressionElement<Float64> element;
    UInt16 type;
    Float64 x;

    _stack.Clear();
    _result = 0;
    errno = 0;
    if (size == 0) {
        return 0;
    }
    for (i = 0; i < size; ++i) {
        element = expression.Get(i);
        type = element.GetType();
        switch (type) {
            case ExpressionElement<Float64>::TYPE_VALUE:
                static_throw(_stack.Push(element.GetValue()));
                break;
            case ExpressionElement<Float64>::TYPE_PLUS:
                Plus();
                break;
            case ExpressionElement<Float64>::TYPE_MINUS:
                Minus();
                break;
            case ExpressionElement<Float64>::TYPE_ADD:
                Add();
                break;
            case ExpressionElement<Float64>::TYPE_SUB:
                Sub();
                break;
            case ExpressionElement<Float64>::TYPE_MUL:
                Mul();
                break;
            case ExpressionElement<Float64>::TYPE_DIV:
                Div();
                break;
            case ExpressionElement<Float64>::TYPE_MOD:
                Mod();
                break;
            case ExpressionElement<Float64>::TYPE_FACTORIAL:
                Factorial();
                break;
            case ExpressionElement<Float64>::TYPE_DFACTORIAL:
                DoubleFactorial();
                break;
            case ExpressionElement<Float64>::TYPE_NPR:
                nPr();
                break;
            case ExpressionElement<Float64>::TYPE_NCR:
                nCr();
                break;
            case ExpressionElement<Float64>::TYPE_POW:
                Pow();
                break;
            case ExpressionElement<Float64>::TYPE_EXP:
                Exp();
                break;
            case ExpressionElement<Float64>::TYPE_LN:
                Ln();
                break;
            case ExpressionElement<Float64>::TYPE_LOG:
                Log();
                break;
            case ExpressionElement<Float64>::TYPE_SQRT:
                Sqrt();
                break;
            case ExpressionElement<Float64>::TYPE_CBRT:
                Cbrt();
                break;
            case ExpressionElement<Float64>::TYPE_SIN:
                Sin();
                break;
            case ExpressionElement<Float64>::TYPE_COS:
                Cos();
                break;
            case ExpressionElement<Float64>::TYPE_TAN:
                Tan();
                break;
            case ExpressionElement<Float64>::TYPE_ASIN:
                Asin();
                break;
            case ExpressionElement<Float64>::TYPE_ACOS:
                Acos();
                break;
            case ExpressionElement<Float64>::TYPE_ATAN:
                Atan();
                break;
            case ExpressionElement<Float64>::TYPE_SINH:
                Sinh();
                break;
            case ExpressionElement<Float64>::TYPE_COSH:
                Cosh();
                break;
            case ExpressionElement<Float64>::TYPE_TANH:
                Tanh();
                break;
            case ExpressionElement<Float64>::TYPE_ASINH:
                Asinh();
                break;
            case ExpressionElement<Float64>::TYPE_ACOSH:
                Acosh();
                break;
            case ExpressionElement<Float64>::TYPE_ATANH:
                Atanh();
                break;
            case ExpressionElement<Float64>::TYPE_ABS:
                Abs();
                break;
            case ExpressionElement<Float64>::TYPE_INT:
                Int();
                break;
            case ExpressionElement<Float64>::TYPE_FIX:
                Fix();
                break;
            case ExpressionElement<Float64>::TYPE_GAMMA:
                Gamma();
                break;
            case ExpressionElement<Float64>::TYPE_FRAC:
                Frac();
                break;
            case ExpressionElement<Float64>::TYPE_JN:
                Jn();
                break;
            case ExpressionElement<Float64>::TYPE_YN:
                Yn();
                break;
            default:
                static_throw(Calculator::ERR_SYNTAX);
                break;
        }
        if (errno == EDOM) {
            static_throw(Calculator::ERR_DOMAIN);
        }
        else if (_stack.GetSize() > 0) {
            x = _stack.GetLast();
            if (!SuperMath::Finite(x)) {
                if (SuperMath::Isnan(x)) {
                    static_throw(Calculator::ERR_NAN);
                }
                else if (x == HUGE_VAL) {
                    static_throw(Calculator::ERR_PLUS_INF);
                }
                else {
                    static_throw(Calculator::ERR_MINUS_INF);
                }
            }
        }
        if (static_catch() != SFERR_NO_ERROR) {
            break;
        }
    }
    if (static_catch() == SFERR_NO_ERROR && _stack.GetSize() != 1) {
        static_throw(Calculator::ERR_SYNTAX);
    }
    if (static_catch() == SFERR_NO_ERROR) {
        _result = _stack.Pop();
    }
    return _result;
}

inline Void Sangi<Float64>::CheckParamCount(UInt16 count) const static_throws
{
    if (_stack.GetSize() < count) {
        static_throw(Calculator::ERR_SYNTAX);
    }
}

inline Void Sangi<Float64>::Plus(Void) static_throws
{
    CheckParamCount(1);
}

inline Void Sangi<Float64>::Minus(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(-param));
}

inline Void Sangi<Float64>::Add(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] + param[1]));
}

inline Void Sangi<Float64>::Sub(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] - param[1]));
}

inline Void Sangi<Float64>::Mul(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    static_throw(_stack.Push(param[0] * param[1]));
}

inline Void Sangi<Float64>::Div(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    if (param[1] == 0) {
        static_throw(Calculator::ERR_ZERO_DIVISION);
    }
    else {
        static_throw(_stack.Push(param[0] / param[1]));
    }
}

inline Void Sangi<Float64>::Mod(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    if (param[1] == 0) {
        static_throw(Calculator::ERR_ZERO_DIVISION);
    }
    else {
        static_throw(_stack.Push(SuperMath::Fmod(param[0], param[1])));
    }
}

inline Void Sangi<Float64>::Factorial(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop() + 1.0), t(SuperMath::Fabs(param));
    if (param <= 0 && t == SuperMath::Floor(t)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Gamma(param)));
    }
}

inline Void Sangi<Float64>::DoubleFactorial(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), t;
    UInt32 n(0);
    if (param < 0.0) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        if (param + 2.0 > static_cast<Float64>(UINT32_MAXIMUM)) {
            static_throw(Calculator::ERR_DOMAIN);
        }
        else {
            t = SuperMath::Floor(param);
            n = static_cast<UInt32>(t);
            if (t != param) {
                static_throw(Calculator::ERR_DOMAIN);
            }
        }
    }
    if (static_catch() == SFERR_NO_ERROR) {
        static_throw(_stack.Push(SuperMath::DoubleFactorial(n)));
    }
}

inline Void Sangi<Float64>::Pow(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2], t;
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    if (param[0] == 0 && param[1] <= 0) {
        static_throw(Calculator::ERR_DOMAIN);
        return;
    }
    if (param[0] < 0) {
        t = SuperMath::Fabs(param[1]);
        if (t != SuperMath::Floor(t)) {
            static_throw(Calculator::ERR_DOMAIN);
            return;
        }
    }
    static_throw(_stack.Push(SuperMath::Pow(param[0], param[1])));
}

inline Void Sangi<Float64>::Exp(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(SuperMath::Exp(param)));
}

inline Void Sangi<Float64>::Ln(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    if (param <= 0) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Log(param)));
    }
}

inline Void Sangi<Float64>::Log(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    if (param <= 0) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Log10(param)));
    }
}

inline Void Sangi<Float64>::Sqrt(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    if (param < 0) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Sqrt(param)));
    }
}

inline Void Sangi<Float64>::Cbrt(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(SuperMath::Cbrt(param)));
}

inline Void Sangi<Float64>::Sin(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), x(0.0);
    switch (_angle) {
        case ANGLEMODE_RAD:
            x = param;
            break;
        case ANGLEMODE_DEG:
            x = SuperMath::FromDegToRad(param);
            break;
        case ANGLEMODE_GRD:
            x = SuperMath::FromGradToRad(param);
            break;
    }
    static_throw(_stack.Push(SuperMath::Sin(x)));
}

inline Void Sangi<Float64>::Cos(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), x(0.0);
    switch (_angle) {
        case ANGLEMODE_RAD:
            x = param;
            break;
        case ANGLEMODE_DEG:
            x = SuperMath::FromDegToRad(param);
            break;
        case ANGLEMODE_GRD:
            x = SuperMath::FromGradToRad(param);
            break;
    }
    static_throw(_stack.Push(SuperMath::Cos(x)));
}

inline Void Sangi<Float64>::Tan(Void) static_throws
{
    Float64Const pi(3.14159265358979323846);

    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), x(0.0);
    switch (_angle) {
        case ANGLEMODE_RAD:
            x = param;
            break;
        case ANGLEMODE_DEG:
            x = SuperMath::FromDegToRad(param);
            break;
        case ANGLEMODE_GRD:
            x = SuperMath::FromGradToRad(param);
            break;
    }
    Float64Const n(SuperMath::Fabs(x / pi - 0.5));
    if (n == SuperMath::Floor(n)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Tan(x)));
    }
}

inline Void Sangi<Float64>::Asin(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), x, y;
    if (param < -1 || param > 1) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        x = SuperMath::Asin(param);
        switch (_angle) {
            case ANGLEMODE_RAD:
                y = x;
                break;
            case ANGLEMODE_DEG:
                y = SuperMath::FromRadToDeg(x);
                break;
            case ANGLEMODE_GRD:
                y = SuperMath::FromRadToGrad(x);
                break;
        }
        static_throw(_stack.Push(y));
    }
}

inline Void Sangi<Float64>::Acos(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), x, y;
    if (param < -1 || param > 1) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        x = SuperMath::Acos(param);
        switch (_angle) {
            case ANGLEMODE_RAD:
                y = x;
                break;
            case ANGLEMODE_DEG:
                y = SuperMath::FromRadToDeg(x);
                break;
            case ANGLEMODE_GRD:
                y = SuperMath::FromRadToGrad(x);
                break;
        }
        static_throw(_stack.Push(y));
    }
}

inline Void Sangi<Float64>::Atan(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), x, y;
    x = SuperMath::Atan(param);
    switch (_angle) {
        case ANGLEMODE_RAD:
            y = x;
            break;
        case ANGLEMODE_DEG:
            y = SuperMath::FromRadToDeg(x);
            break;
        case ANGLEMODE_GRD:
            y = SuperMath::FromRadToGrad(x);
            break;
    }
    static_throw(_stack.Push(y));
}

inline Void Sangi<Float64>::Sinh(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(SuperMath::Sinh(param)));
}

inline Void Sangi<Float64>::Cosh(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(SuperMath::Cosh(param)));
}

inline Void Sangi<Float64>::Tanh(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(SuperMath::Tanh(param)));
}

inline Void Sangi<Float64>::Asinh(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(SuperMath::Asinh(param)));
}

inline Void Sangi<Float64>::Acosh(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    if (param < 1) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Acosh(param)));
    }
}

inline Void Sangi<Float64>::Atanh(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    if (param <= -1 || param >= 1) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Atanh(param)));
    }
}

inline Void Sangi<Float64>::Abs(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(SuperMath::Fabs(param)));
}

inline Void Sangi<Float64>::Int(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop());
    static_throw(_stack.Push(SuperMath::Floor(param)));
}

inline Void Sangi<Float64>::Fix(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), x;
    if (param >= 0) {
        x = SuperMath::Floor(param);
    }
    else {
        x = -SuperMath::Floor(-param);
    }
    static_throw(_stack.Push(x));
}

inline Void Sangi<Float64>::Gamma(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), t(SuperMath::Fabs(param));
    if (param <= 0 && t == SuperMath::Floor(t)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Gamma(param)));
    }
}

inline Void Sangi<Float64>::nPr(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    if (param[0] < 0 || param[0] < 0 || param[0] < param[1]) {
        static_throw(Calculator::ERR_DOMAIN);
        return;
    }
    Float64 n[2] = {SuperMath::Floor(param[0]), SuperMath::Floor(param[1])};
    if (param[0] != n[0] || param[1] != n[1]) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    if (param[0] > static_cast<Float64>(UINT32_MAXIMUM) || param[1] > static_cast<Float64>(UINT32_MAXIMUM)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Permutation(static_cast<UInt32>(n[0]), static_cast<UInt32>(n[1]))));
    }
}

inline Void Sangi<Float64>::nCr(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2];
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    if (param[0] < 0 || param[0] < 0 || param[0] < param[1]) {
        static_throw(Calculator::ERR_DOMAIN);
        return;
    }
    Float64 n[2] = {SuperMath::Floor(param[0]), SuperMath::Floor(param[1])};
    if (param[0] != n[0] || param[1] != n[1]) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    if (param[0] > static_cast<Float64>(UINT32_MAXIMUM) || param[1] > static_cast<Float64>(UINT32_MAXIMUM)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Combination(static_cast<UInt32>(n[0]), static_cast<UInt32>(n[1]))));
    }
}

inline Void Sangi<Float64>::Frac(Void) static_throws
{
    CheckParamCount(1);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param(_stack.Pop()), x;
    if (param >= 0) {
        x = param - SuperMath::Floor(param);
    }
    else {
        x = - ((-param) - SuperMath::Floor(-param));
    }
    static_throw(_stack.Push(x));
}

inline Void Sangi<Float64>::Jn(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2], n;
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    n = SuperMath::Fabs(param[0]);
    if (n != SuperMath::Floor(n)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else if (n > static_cast<Float64>(SINT32_MAXIMUM) || n < static_cast<Float64>(SINT32_MINIMUM)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Jn(static_cast<SInt32>(param[0]), param[1])));
    }
}

inline Void Sangi<Float64>::Yn(Void) static_throws
{
    CheckParamCount(2);
    if (static_catch() != SFERR_NO_ERROR) {
        return;
    }

    Float64 param[2], n;
    param[1] = _stack.Pop();
    param[0] = _stack.Pop();
    n = SuperMath::Fabs(param[0]);
    if (n != SuperMath::Floor(n)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else if (n > static_cast<Float64>(SINT32_MAXIMUM) || n < static_cast<Float64>(SINT32_MINIMUM)) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else if (param[1] <= 0) {
        static_throw(Calculator::ERR_DOMAIN);
    }
    else {
        static_throw(_stack.Push(SuperMath::Yn(static_cast<SInt32>(param[0]), param[1])));
    }
}

inline Float64 Sangi<Float64>::GetResult(Void) const
{
    return _result;
}

#endif // __SOPHIACRADLE_SANGI_HPP //
